# Copyright 2013 Canonical Ltd. This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import unicode_literals

class UnexpectedApiError(Exception):
    """An unexpected client error."""

    def __init__(self, response=None, msg="", json=None):
        self.response = response
        self.json_body = json
        if response:
            msg = "%s : %s - %s" % (response.status_code, response.text, msg)
        super(UnexpectedApiError, self).__init__(msg)


class ServerError(UnexpectedApiError):
    """An unexpected 5xx response."""


class ClientError(UnexpectedApiError):
    """An unexpected 4xx response."""


class ApiException(Exception):
    """An expected/understood 4xx or 5xx response.

    Parse the standard api error reponse format given by SSO.

    """
    def __init__(self, response, body=None, msg=None):
        body = body or {}
        self.response = response
        self.body = body
        self.error_message = body.get('message')
        self.extra = body.get('extra', {})
        if msg is None:
            # code *should* be the same as the error_code attribute
            # but won't be for raising an ApiException directly instead
            # of a subclass - so still fetch it from the payload body
            code = body.get('code')
            msg = "%s: %s" % (response.status_code, code)
            extra = ', '.join('%s: %r' % i for i in self.extra.items())
            if extra:
                msg += ' (%s)' % extra
        super(ApiException, self).__init__(msg)


class InvalidData(ApiException):
    error_code = "INVALID_DATA"
    status_code = 400


class CaptchaRequired(ApiException):
    error_code = "CAPTCHA_REQUIRED"
    status_code = 401


class InvalidCredentials(ApiException):
    error_code = "INVALID_CREDENTIALS"
    status_code = 401


class TwoFactorRequired(ApiException):
    error_code = "TWOFACTOR_REQUIRED"
    status_code = 401


class AccountSuspended(ApiException):
    error_code = "ACCOUNT_SUSPENDED"
    status_code = 403


class AccountDeactivated(ApiException):
    error_code = "ACCOUNT_DEACTIVATED"
    status_code = 403


class AccountLocked(ApiException):
    error_code = "ACCOUNT_LOCKED"
    status_code = 403


class EmailInvalidated(ApiException):
    error_code = "EMAIL_INVALIDATED"
    status_code = 403


class CanNotResetPassword(ApiException):
    error_code = "CAN_NOT_RESET_PASSWORD"
    status_code = 403


class CaptchaFailure(ApiException):
    error_code = "CAPTCHA_FAILURE"
    status_code = 403


class TooManyTokens(ApiException):
    error_code = "TOO_MANY_TOKENS"
    status_code = 403


class TwoFactorFailure(ApiException):
    error_code = "TWOFACTOR_FAILURE"
    status_code = 403


class PasswordPolicyError(ApiException):
    error_code = "PASSWORD_POLICY_ERROR"
    status_code = 403


class ResourceNotFound(ApiException):
    error_code = "RESOURCE_NOT_FOUND"
    status_code = 404


class AlreadyRegistered(ApiException):
    error_code = "ALREADY_REGISTERED"
    status_code = 409


class TooManyRequests(ApiException):
    error_code = "TOO_MANY_REQUESTS"
    status_code = 429


class CaptchaError(ApiException):
    error_code = "CAPTCHA_ERROR"
    status_code = 502



# Copyright 2013 Canonical Ltd. This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).
from .errors import (
    AccountDeactivated,
    AccountLocked,
    AccountSuspended,
    AlreadyRegistered,
    ApiException,
    CanNotResetPassword,
    CaptchaError,
    CaptchaFailure,
    CaptchaRequired,
    ClientError,
    EmailInvalidated,
    InvalidCredentials,
    InvalidData,
    PasswordPolicyError,
    TooManyTokens,
    TwoFactorFailure,
    ResourceNotFound,
    ServerError,
    TooManyRequests,
    TwoFactorRequired,
    UnexpectedApiError,
)
from .client import V2ApiClient



# Copyright 2013 Canonical Ltd. This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import unicode_literals

import functools
import json

import requests

from ssoclient.v2 import errors

JSON_MIME_TYPE = 'application/json'
ERRORS = {
    e.error_code: e for e in [
        errors.AccountDeactivated,
        errors.AccountLocked,
        errors.AlreadyRegistered,
        errors.AccountSuspended,
        errors.CanNotResetPassword,
        errors.CaptchaError,
        errors.CaptchaFailure,
        errors.CaptchaRequired,
        errors.EmailInvalidated,
        errors.InvalidCredentials,
        errors.InvalidData,
        errors.PasswordPolicyError,
        errors.ResourceNotFound,
        errors.TooManyRequests,
        errors.TooManyTokens,
        errors.TwoFactorFailure,
        errors.TwoFactorRequired,
    ]
}


class V2ApiClientResponse(object):
    """A successful response from an API call to the v2 SSO API."""

    def __init__(self, status_code, content):
        super(V2ApiClientResponse, self).__init__()
        self.status_code = status_code
        self.content = content
        self.ok = True  # backwards compat with requests' Response

    def json(self):
        """Return this instance's content, available for backwards compat."""
        return self.content


def process_response(func):
    """Decorator to parse 'func' responses and manage errors."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        """Wrapper for func."""

        response = func(*args, **kwargs)
        try:
            json_body = response.json()
        except (ValueError, TypeError):
            # simplejson raises JSONDecodeError for invalid json
            # json raises ValueError. JSONDecodeError is a subclass
            # of ValueError - so this catches either.
            json_body = {}

        if response.ok:
            # return the json-decoded value and the status code
            return V2ApiClientResponse(
                status_code=response.status_code, content=json_body)

        # from now on, just error management
        code = json_body.get('code')
        exc = ERRORS.get(code)
        if exc is not None:
            # raise a specific exception
            raise exc(response, json_body)

        if code:
            msg = "Unknown error code '%s' in response" % code
        else:
            msg = "No error code in response"

        if response.content:
            msg += ' (%r)' % response.content

        if response.status_code < 500:
            exc = errors.ClientError
        else:
            exc = errors.ServerError
        raise exc(response, msg, json_body)

    return wrapper


class ApiSession(requests.Session):
    """An SSO api specfic Session

    Adds support for a url endpoint, 500 exceptions, JSON request body
    handling and retries.
    """

    def __init__(self, endpoint):
        super(ApiSession, self).__init__()
        self.endpoint = endpoint.rstrip('/') + '/'
        # sent with every request
        self.headers['Accept'] = JSON_MIME_TYPE
        for protocol in ('http://', 'https://'):
            self.mount(protocol, requests.adapters.HTTPAdapter(max_retries=3))

    @process_response
    def request(self, method, url, **kwargs):
        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        if 'data' in kwargs:
            kwargs['data'] = json.dumps(kwargs['data'])
            kwargs['headers']['Content-Type'] = JSON_MIME_TYPE
        url = self.endpoint + url.lstrip('/')
        response = super(ApiSession, self).request(method, url, **kwargs)
        return response



# Copyright 2013 Canonical Ltd. This software is licensed under
# the GNU Affero General Public License version 3 (see the file
# LICENSE).

from __future__ import unicode_literals

try:
    str = unicode
except NameError:
    pass  # Forward compatibility with Py3k

import logging

from datetime import datetime

from requests_oauthlib import OAuth1

from ssoclient.v2.http import ApiSession
from ssoclient.v2 import errors


logger = logging.getLogger(__name__)


def datetime_from_string(value):
    """Parse a string value representing a date in isoformat."""
    try:
        result = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        result = datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
    return result


def parse_datetimes(value):
    """Recursively look for dates and try to parse them into datetimes."""
    assert isinstance(value, dict)
    result = value.copy()
    for k, v in value.items():
        if isinstance(v, dict):
            result[k] = parse_datetimes(v)
        elif isinstance(v, list):
            result[k] = [parse_datetimes(i) for i in v]
        elif 'date' in k:
            result[k] = datetime_from_string(v)
    return result


class V2ApiClient(object):
    """High-level client for theV2.0 API SSO resources."""

    def __init__(self, endpoint):
        self.session = ApiSession(endpoint)

    def _unicode_credentials(self, credentials):
        # if openid and credentials come directly from a call to client.login
        # then whether they are unicode or byte-strings depends on which
        # json library is in use.
        # oauthlib requires them to be unicode - so we coerce to be sure.
        if credentials is not None:
            consumer_key = str(credentials.get('consumer_key', ''))
            consumer_secret = str(credentials.get('consumer_secret', ''))
            token_key = str(credentials.get('token_key', ''))
            token_secret = str(credentials.get('token_secret', ''))
            oauth = OAuth1(
                consumer_key,
                consumer_secret,
                token_key, token_secret,
            )
        else:
            oauth = None
        return oauth

    def _merge(self, data, extra):
        """Allow data to passed to functions by keyword or by dict."""
        if data:
            data.update(extra)
        else:
            data = extra
        return data

    def register(self, data=None, **kwargs):
        response = self.session.post(
            '/accounts', data=self._merge(data, kwargs))
        result = parse_datetimes(response.content)
        return result

    def login(self, data=None, **kwargs):
        response = self.session.post(
            '/tokens/oauth', data=self._merge(data, kwargs))
        result = parse_datetimes(response.content)
        return result

    def get_or_create_account(self, token, **kwargs):
        oauth = self._unicode_credentials(token)
        response = self.session.post(
            '/accounts', data=kwargs, auth=oauth)
        result = parse_datetimes(response.content)
        created = response.status_code == 201
        return result, created

    def account_details(self, openid, token=None, expand=False):
        openid = str(openid)
        oauth = self._unicode_credentials(token)
        url = '/accounts/%s?expand=%s' % (openid, str(expand).lower())

        response = self.session.get(url, auth=oauth)
        result = parse_datetimes(response.content)
        return result

    def email_delete(self, email, credentials):
        oauth = self._unicode_credentials(credentials)
        response = self.session.delete('/emails/%s' % email, auth=oauth)
        return response.status_code == 204

    def email_details(self, email, credentials):
        oauth = self._unicode_credentials(credentials)

        response = self.session.get('/emails/%s' % email, auth=oauth)
        result = parse_datetimes(response.content)
        return result

    def token_delete(self, token_key, credentials):
        oauth = self._unicode_credentials(credentials)
        response = self.session.delete(
            '/tokens/oauth/%s' % token_key, auth=oauth)
        return response.status_code == 204

    def token_details(self, token_key, credentials):
        oauth = self._unicode_credentials(credentials)

        response = self.session.get('/tokens/oauth/%s' % token_key, auth=oauth)
        result = parse_datetimes(response.content)
        return result

    def validate_request(self, data=None, **kwargs):
        """Validate an OAuth signature.

        The OAuth signature can be given either as the value of the
        Authorization header, or as a query string with the OAuth information
        in it. Expected parameters are:

        * 'http_url'
        * 'http_method'
        * 'authorization' and/or 'query_string' (at least one)

        Return a dictionary with a 'is_valid' field that indicates whether the
        given OAuth signature is valid or not.

        """
        data = self._merge(data, kwargs)
        http_url = data.get('http_url', '')
        if not isinstance(http_url, str):
            try:
                data['http_url'] = http_url.decode('utf-8')
            except UnicodeError:
                msg = ('Given http_url %r can not be utf-8 decoded' % http_url)
                logger.error(
                    'validate_request: %s, raising ClientError.', msg)
                raise errors.ClientError(msg=msg)

        response = self.session.post('/requests/validate', data=data)
        return response.content

    def request_password_reset(self, email, token=None):
        response = self.session.post(
            '/tokens/password', data=dict(email=email, token=token))
        return response.content
